---
title: llarp/rpc/rpc_request_decorators.hpp

---

# llarp/rpc/rpc_request_decorators.hpp



## Namespaces

| Name           |
| -------------- |
| **[tools](Namespaces/namespacetools.md)**  |
| **[llarp](Namespaces/namespacellarp.md)** <br>[crypto.hpp] |
| **[llarp::rpc](Namespaces/namespacellarp_1_1rpc.md)**  |

## Classes

|                | Name           |
| -------------- | -------------- |
| struct | **[tools::type_list](Classes/structtools_1_1type__list.md)**  |
| struct | **[llarp::rpc::RPCRequest](Classes/structllarp_1_1rpc_1_1RPCRequest.md)**  |
| struct | **[llarp::rpc::NoArgs](Classes/structllarp_1_1rpc_1_1NoArgs.md)**  |
| struct | **[llarp::rpc::Deferred](Classes/structllarp_1_1rpc_1_1Deferred.md)**  |
| struct | **[llarp::rpc::Immediate](Classes/structllarp_1_1rpc_1_1Immediate.md)**  |




## Source code

```cpp
#pragma once

#include "json_binary_proxy.hpp"
#include <string_view>
#include <llarp/rpc/endpoints.hpp>
#include <llarp/config/config.hpp>
#include <oxenmq/oxenmq.h>
#include <oxenmq/address.h>
#include <oxen/log/omq_logger.hpp>


namespace tools
{
  //  Type wrapper that contains an arbitrary list of types.
  template <typename...> struct type_list {};
} // namespace llarp


namespace llarp::rpc
{
  //  Returns a constexpr std::array of string_views from an arbitrary list of string literals
  //  Used to specify RPC names as:
  //  static constexpr auto names() { return NAMES("primary_name", "some_alias"); }
  template <size_t... N>
  constexpr std::array<std::string_view, sizeof...(N)> NAMES(const char (&...names)[N]) {
    static_assert(sizeof...(N) > 0, "RPC command must have at least one name");
    return {std::string_view{names, N-1}...};
  }


  //  Base class that all RPC requests will expand for each endpoint type
  struct RPCRequest {
    private:
      bool bt = false;

    public:
      //  Returns true if response is bt-encoded, and false for json
      //  Note: do not set value
      bool is_bt() const { return bt; }

      // Callable method to indicate request is bt-encoded
      void set_bt() { 
        bt = true; 
        response_b64.format = tools::json_binary_proxy::fmt::bt;
        response_hex.format = tools::json_binary_proxy::fmt::bt;
      }

      //  Response Data:
      //  bt-encoded are converted in real-time
      //    - bool becomes 0 or 1
      //    - key:value where value == null are omitted
      //    - other nulls will raise an exception if found in json
      //    - no doubles
      //      - to store doubles: encode bt in endpoint-specific way
      //    - binary strings will fail json serialization; caller must
      //    
      //        std::string binary = some_binary_data();
      //        request.response["binary_value"] = is_bt ? binary : oxenmq::to_hex(binary)
      //    
      nlohmann::json response;

      //  Proxy Object:
      //  Sets binary data in "response"
      //    - if return type is json, encodes as hex
      //    - if return type is bt, then binary is untouched
      //  
      //  Usage:
      //    std::string data = "abc";
      //    request.response_hex["foo"]["bar"] = data; // json: "616263", bt: "abc"
      //
      tools::json_binary_proxy response_hex{response, tools::json_binary_proxy::fmt::hex};

      //  Proxy Object:
      //  Encodes binary data as base_64 for json-encoded responses, leaves as binary for bt-encoded responses
      //  
      //    Usage:
      //      std::string data = "abc"
      //      request.response_b64["foo"]["bar"] = data; json: "YWJj", bt: "abc"
      //      
      tools::json_binary_proxy response_b64{response, tools::json_binary_proxy::fmt::base64};
  };

  //  Tag types that are inherited to set RPC endpoint properties

  //  RPC call wil take no input arguments
  //    Parameter dict can be passed, but will be ignored
  struct NoArgs : virtual RPCRequest {};

  //  RPC call will be executed in the logic thread cycle
  struct Deferred : virtual RPCRequest {};

  // RPC call will be executed immediately
  struct Immediate : virtual RPCRequest {};

}  // namespace llarp::rpc
```


-------------------------------

Updated on 2023-01-13 at 22:44:39 +0000
